unit ugenere;

{}
{}                                                                        {}
{}                               INTERFACE                                {}
{}                                                                        {}
{}

{  UNITES  }
uses crt;

{  CONSTANTES  }
const
     largeur_max  = 30; {>2}
     longueur_max = 50; {>2}

{  TYPES  }

type
    Tdiror = (H,D,B,G); {les 4 directions}
    Tposition = record
                abs:integer;
                ord:integer;
                diror:Tdiror
                end;
{matrice}
    Tcase = array [H..G] of boolean; {vrai si ouverture dans la direction}
    Tmat = array [1..longueur_max,1..largeur_max] of Tcase;

{  VARIABLES  }

var
   mat:Tmat;
   longueur,largeur:byte;

procedure generer_laby;


{}
{}                                                                        {}
{}                             IMPLEMENTATION                             {}
{}                                                                        {}
{}

{  TYPES  }

type

{chemin}
    Tpt_chemin = ^Tcellule;
    Tcellule = record
               abs_chem:integer;
               ord_chem:integer;
               suiv:Tpt_chemin;
               end;

{reste du labyrinthe}
    Tpt_arbre = ^Tarbre;
    Tdirsuiv = (Ga,F,Dr);
    Ttab_dir = array [Ga..Dr] of Tpt_arbre;
    Tarbre = record
             abs_arbre:integer;
             ord_arbre:integer;
             diror_arbre:Tdiror;
             tab_dir:Ttab_dir;
             end;

{  VARIABLES  }

var
   chemin:Tpt_chemin;


{ CREATION DU CHEMIN }

{****************************************************************************}
function cellules_differentes (pt1,pt2:Tpt_chemin):boolean;
begin
   cellules_differentes:= ( pt1^.abs_chem <> pt2^.abs_chem ) or
                          ( pt1^.ord_chem <> pt2^.ord_chem );
end;
{****************************************************************************}

{****************************************************************************}
function chemin_cree (pt1,pt2:Tpt_chemin):boolean;
var
   memes_cellules:boolean;
   pt_cour:Tpt_chemin;
   abscisse,ordonnee:integer;
begin
   memes_cellules:=not cellules_differentes (pt1,pt2);
   pt_cour:=pt2;
   while (pt_cour<>nil) and (not memes_cellules) do
   begin
      pt_cour:=pt_cour^.suiv;
      memes_cellules:=not cellules_differentes (pt1,pt_cour);
   end;
   chemin_cree:=memes_cellules;
end;
{****************************************************************************}

{****************************************************************************}
procedure supprimer_boucle (var pt:Tpt_chemin);
var
   pt_cour:Tpt_chemin;
begin
   pt_cour:=pt^.suiv;
   while cellules_differentes(pt_cour,pt) do pt_cour:=pt_cour^.suiv;
   pt:=pt_cour;
end;
{****************************************************************************}

{****************************************************************************}
procedure inserer_en_tete (var nouv,Tdl:Tpt_chemin);
begin
   nouv^.suiv:=Tdl;
   Tdl:=nouv;
end;
{****************************************************************************}

{****************************************************************************}
function nouvelle_cellule (abscisse,ordonnee:integer):Tpt_chemin;
var
   nouv:Tpt_chemin;
begin
   new(nouv);
   with nouv^ do
   begin
      abs_chem:=abscisse;
      ord_chem:=ordonnee;
      suiv:=nil;
   end;
   nouvelle_cellule:=nouv;
end;
{****************************************************************************}

{****************************************************************************}
function nb_cellules (pt:Tpt_chemin):integer;
var
   pt_cour:Tpt_chemin;
   nb:integer;
begin
   pt_cour:=pt;
   nb:=0;
   while (pt_cour<>nil) do
   begin
      inc(nb);
      pt_cour:=pt_cour^.suiv;
   end;
   nb_cellules:=nb;
end;
{****************************************************************************}

{****************************************************************************}
procedure choisir_case(pt:Tpt_chemin;var pt_choisi:Tpt_chemin);
var
   r,i:integer;
begin
   r:=random(nb_cellules(pt));
   pt_choisi:=pt;
   for i:=1 to r do pt_choisi:=pt_choisi^.suiv;
   pt_choisi:=nouvelle_cellule(pt_choisi^.abs_chem,pt_choisi^.ord_chem);
end;
{****************************************************************************}

{****************************************************************************}
function elle_est_libre (pt:Tpt_chemin;abs,ord:integer):boolean;
var
   libre:boolean;
   pt_cour:Tpt_chemin;
begin
   libre:=true;
   pt_cour:=pt;
   while (pt_cour<>nil) and (libre) do
   begin
      libre:= (pt_cour^.abs_chem<>abs) or (pt_cour^.ord_chem<>ord);
      pt_cour:=pt_cour^.suiv;
   end;
   elle_est_libre := libre;
end;
{****************************************************************************}

{****************************************************************************}
function elle_existe (abs,ord:integer):boolean;
begin
   elle_existe := (abs>0) and (abs<=longueur) and (ord>0) and (ord<=largeur);
end;
{****************************************************************************}

{****************************************************************************}
procedure tester_la_case (pt:Tpt_chemin;var libres,occuppees:Tpt_chemin;
                          abs,ord:integer);
var
   nouv:Tpt_chemin;
begin
   if elle_existe (abs,ord) then
   begin
      nouv:=nouvelle_cellule (abs,ord);
      if elle_est_libre (pt,abs,ord)
         then inserer_en_tete (nouv,libres)
         else inserer_en_tete (nouv,occuppees);
   end;
end;
{****************************************************************************}

{****************************************************************************}
procedure cases_prolongeables (pt:Tpt_chemin;var libres,occuppees:Tpt_chemin);
var
   abscisse,ordonnee:integer;
begin
   abscisse:=pt^.abs_chem;
   ordonnee:=pt^.ord_chem;
   dec(ordonnee);
   tester_la_case(pt,libres,occuppees,abscisse,ordonnee);
   inc(ordonnee,2);
   tester_la_case(pt,libres,occuppees,abscisse,ordonnee);
   dec(ordonnee);
   dec(abscisse);
   tester_la_case(pt,libres,occuppees,abscisse,ordonnee);
   inc(abscisse,2);
   tester_la_case(pt,libres,occuppees,abscisse,ordonnee);
end;
{****************************************************************************}

{****************************************************************************}
procedure prolonger (var pt:Tpt_chemin);
var
   libres,occuppees,pt_inser:Tpt_chemin;
begin
   libres:=nil;
   occuppees:=nil;
   cases_prolongeables(pt,libres,occuppees);
{   afficher_les_listes (pt);
   afficher_les_listes (libres);
   afficher_les_listes (occuppees);}
   if (libres<>nil) then
   begin
      choisir_case(libres,pt_inser);
      inserer_en_tete(pt_inser,pt);
   end else
   begin
      choisir_case(occuppees,pt_inser);
      inserer_en_tete(pt_inser,pt);
      supprimer_boucle(pt);
{   afficher_les_listes (pt);
   writeln;}
   end;
end;
{****************************************************************************}

{****************************************************************************}
procedure enleve_le_surplus (var liste_surplus:Tpt_chemin;coordonnees:Tpt_chemin);
begin
   while cellules_differentes(liste_surplus,coordonnees) do
         liste_surplus:=liste_surplus^.suiv;
end;
{****************************************************************************}

{****************************************************************************}
procedure concatener (var depart,arrivee:Tpt_chemin);
var
   nouv:Tpt_chemin;
begin
   depart:=depart^.suiv;
   while (depart<>nil) do
   begin
      nouv:=nouvelle_cellule (depart^.abs_chem,depart^.ord_chem);
      inserer_en_tete(nouv,arrivee);
      depart:=depart^.suiv;
   end;
   depart:=arrivee;
end;
{****************************************************************************}

{****************************************************************************}
procedure creer_le_chemin(var depart:Tpt_chemin);
var
   arrivee:Tpt_chemin;
   fini_dep,fini_arr:boolean;
begin
   depart:=nouvelle_cellule(1,1);
   arrivee:=nouvelle_cellule(longueur,largeur);
   repeat
      prolonger(depart);
      fini_dep:=chemin_cree(depart,arrivee);
      if not fini_dep then
      begin
         prolonger(arrivee);
         fini_arr:=chemin_cree(arrivee,depart);
      end;
   until (fini_dep) or (fini_arr);
   if fini_dep then enleve_le_surplus (arrivee,depart)
               else enleve_le_surplus (depart,arrivee);
   concatener (depart,arrivee);
end;
{****************************************************************************}


{ RESTE DU LABYRINTHE }

{****************************************************************************}
procedure calcul_variations (dirsuiv:Tdirsuiv;diror:Tdiror;var dx,dy:integer;
                             var dir_or:Tdiror);
begin
   dx:=0;
   dy:=0;
   case diror of
        H:case dirsuiv of
               Ga:inc(dx);
               F :inc(dy);
               Dr:dec(dx);
          end;
        B:case dirsuiv of
               Ga:dec(dx);
               F :dec(dy);
               Dr:inc(dx);
          end;
        G:case dirsuiv of
               Ga:dec(dy);
               F :inc(dx);
               Dr:inc(dy);
          end;
        H:case dirsuiv of
               Ga:inc(dy);
               F :dec(dx);
               Dr:dec(dy);
          end;
   end; {of case}
   case dirsuiv of
        Ga:if diror=H then dir_or:=G
                      else dir_or:=pred(diror);
        F :dir_or:=diror;
        Dr:if diror=G then dir_or:=H
                      else dir_or:=succ(diror);
   end;
end;
{****************************************************************************}

{****************************************************************************}
function case_isolee (case_testee:Tcase):boolean;
var
   isolee:boolean;
   diror:Tdiror;
begin
   isolee:=true;
   for diror:= H to G do isolee := isolee and case_testee [diror];
   case_isolee:=isolee;
end;
{****************************************************************************}

{****************************************************************************}
function nouvelle_cellule_arbre (abs,ord:integer;dir_or:Tdiror):Tpt_arbre;
var
   nouv:Tpt_arbre;
   dirsuiv:Tdirsuiv;
begin
   new(nouv);
   with nouv^ do
   begin
      abs_arbre:=abs;
      ord_arbre:=ord;
      diror_arbre:=dir_or;
      for dirsuiv:=Ga to Dr do tab_dir[dirsuiv]:=nil;
   end;
   nouvelle_cellule_arbre:=nouv;
end;
{****************************************************************************}

{****************************************************************************}
procedure casser_le_mur (var mat:Tmat;abs,ord,dx,dy:integer);
begin
   case dx of
        -1:begin
              mat[abs,ord][G]:=false;
              mat[abs-1,ord][D]:=false;
           end;
         0:case dy of
                -1:begin
                      mat[abs,ord][H]:=false;
                      mat[abs,ord-1][B]:=false;
                   end;
                 1:begin
                      mat[abs,ord][B]:=false;
                      mat[abs,ord+1][H]:=false;
                   end;
           end; {of case}
         1:begin
              mat[abs,ord][D]:=false;
              mat[abs+1,ord][G]:=false;
           end;
   end; {of case}
end;
{****************************************************************************}

{****************************************************************************}
procedure tracer_le_chemin (var mat:Tmat;chemin:Tpt_chemin);
var
   pt_cour:Tpt_chemin;
   dx,dy:integer;
begin
   pt_cour:=chemin;
   while (pt_cour^.suiv<>nil) do
   begin
      dx:=pt_cour^.suiv^.abs_chem-pt_cour^.abs_chem;
      dy:=pt_cour^.suiv^.ord_chem-pt_cour^.ord_chem;
      casser_le_mur(mat,pt_cour^.abs_chem,pt_cour^.ord_chem,dx,dy);
      pt_cour:=pt_cour^.suiv;
   end;
end;
{****************************************************************************}

{****************************************************************************}
procedure init_matrice (var mat:Tmat);
var
   long,larg:integer;
   diror:Tdiror;
begin
   for larg:=1 to largeur do
       for long:=1 to longueur do
           for diror:=H to G do mat[long,larg][diror]:=true;
{   mat[1,1][G]:=false;
   mat[longueur,largeur][D]:=false;}
end;
{****************************************************************************}

{****************************************************************************}
function avancer (dirsuiv:Tdirsuiv;pta:Tpt_arbre;chance:integer):boolean;
var
   avance:boolean;
   dx,dy,r:integer;
   diror_inutile:Tdiror;
begin
   chance:=2*chance;if chance=2 then chance:=3;
   avance:=false;
   r:=random(6);
   if r<chance then
   begin
      calcul_variations(dirsuiv,pta^.diror_arbre,dx,dy,diror_inutile);
      if elle_existe(pta^.abs_arbre+dx,pta^.ord_arbre+dy) then
         avance:=case_isolee(mat[pta^.abs_arbre+dx,pta^.ord_arbre+dy]);
   end;
   avancer:=avance;
end;
{****************************************************************************}

{****************************************************************************}
procedure faire_arbre (var mat:Tmat;var pta:Tpt_arbre;chance:integer);
var
   dirsuiv:Tdirsuiv;
   dx,dy:integer;
   diror:Tdiror;
begin
   for dirsuiv:=Ga to Dr do
       if avancer (dirsuiv,pta,chance) then
       begin
          calcul_variations(dirsuiv,pta^.diror_arbre,dx,dy,diror);
          pta^.tab_dir[dirsuiv] := nouvelle_cellule_arbre (pta^.abs_arbre+dx,
                                                           pta^.ord_arbre+dy,
                                                           diror);
          casser_le_mur (mat,pta^.abs_arbre,pta^.ord_arbre,dx,dy);
          faire_arbre (mat,pta^.tab_dir[dirsuiv],chance);
       end;
end;
{****************************************************************************}

{****************************************************************************}
procedure faire_case (var mat:Tmat;pt:Tpt_chemin;diror:Tdiror;chance:integer);
var
   Tda:Tpt_arbre;
   r,dx,dy:integer;
   dir_or:Tdiror;
begin
   calcul_variations (F,diror,dx,dy,dir_or);
   if (elle_existe(pt^.abs_chem+dx,pt^.ord_chem+dy)) then
      if case_isolee (mat[pt^.abs_chem+dx,pt^.ord_chem+dy]) then
      begin
         if chance=2 then r:=0 else r:=random(3);
         if r=0 then
         begin
            Tda:=nouvelle_cellule_arbre (pt^.abs_chem+dx,pt^.ord_chem+dy,dir_or);
            casser_le_mur (mat,pt^.abs_chem,pt^.ord_chem,dx,dy);
            faire_arbre (mat,Tda,chance);
         end;
      end;
end;
{****************************************************************************}

{****************************************************************************}
procedure faire_le_chemin (var mat:Tmat;chemin:Tpt_chemin;chance:integer);
var
   pt_cour:Tpt_chemin;
begin
   pt_cour:=chemin;
   while (pt_cour<>nil) do
   begin
      faire_case (mat,pt_cour,H,chance);
      faire_case (mat,pt_cour,B,chance);
      faire_case (mat,pt_cour,G,chance);
      faire_case (mat,pt_cour,D,chance);
      pt_cour:=pt_cour^.suiv;
   end;
end;
{****************************************************************************}

{****************************************************************************}
procedure desisoler_la_case (var mat:Tmat;abs,ord:integer);
var
   pt,nouv:Tpt_chemin;
   i,r:integer;
begin
   pt:=nil;
   if elle_existe (abs,ord-1) then
      if not case_isolee (mat[abs,ord-1]) then
      begin
         nouv:=nouvelle_cellule(abs,ord-1);
         inserer_en_tete (nouv,pt);
      end;
   if elle_existe (abs,ord+1) then
      if not case_isolee (mat[abs,ord+1]) then
      begin
         nouv:=nouvelle_cellule(abs,ord+1);
         inserer_en_tete (nouv,pt);
      end;
   if elle_existe (abs-1,ord) then
      if not case_isolee (mat[abs-1,ord]) then
      begin
         nouv:=nouvelle_cellule(abs-1,ord);
         inserer_en_tete (nouv,pt);
      end;
   if elle_existe (abs+1,ord) then
      if not case_isolee (mat[abs+1,ord]) then
      begin
         nouv:=nouvelle_cellule(abs+1,ord);
         inserer_en_tete (nouv,pt);
      end;
   r:=random(nb_cellules(pt));
   for i:=1 to r do pt:=pt^.suiv;
   casser_le_mur (mat,abs,ord,pt^.abs_chem-abs,pt^.ord_chem-ord);
end;
{****************************************************************************}

{****************************************************************************}
procedure desisoler_les_cases (var mat:Tmat);
var
   larg,long:integer;
begin
   for larg:=1 to largeur do
       for long:=1 to longueur do
           if case_isolee (mat[long,larg]) then
              desisoler_la_case (mat,long,larg);
end;
{****************************************************************************}

{****************************************************************************}
procedure finir_laby (var mat:Tmat;chemin:Tpt_chemin);
begin
   {1 chance sur 3 de commencer un arbre et 1 sur 2 de le prolonger}
   faire_le_chemin (mat,chemin,1);
   {1 chance sur 1 de commencer un arbre et 2 sur 3 de le prolonger}
   faire_le_chemin (mat,chemin,2);
   {1 chance sur 1 de rendre une case accessible}
   desisoler_les_cases (mat);
end;
{****************************************************************************}

{****************************************************************************}
procedure generer_laby;
begin
  creer_le_chemin (chemin);
  init_matrice (mat);
  tracer_le_chemin (mat,chemin);
  finir_laby (mat,chemin);
end;
{****************************************************************************}

{}
{                                                                          }
{                             INITIALISATIONS                              }
{                                                                          }
{}

BEGIN
  randomize;
END.



